package ru.dbotthepony.mc.otm.client.screen.panels.util

import net.minecraft.client.gui.screens.Screen
import ru.dbotthepony.mc.otm.client.render.RenderGravity
import ru.dbotthepony.mc.otm.client.screen.panels.Dock
import ru.dbotthepony.mc.otm.client.screen.panels.EditablePanel
import ru.dbotthepony.mc.otm.client.screen.panels.slot.AbstractSlotPanel
import ru.dbotthepony.mc.otm.core.collect.filter
import ru.dbotthepony.mc.otm.core.collect.maybe
import kotlin.math.max

open class GridPanel<out S : Screen>(
	screen: S,
	parent: EditablePanel<*>?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = 0f,
	height: Float = 0f,
	columns: Int,
	rows: Int
) : EditablePanel<S>(screen, parent, x, y, width, height) {
	var columns: Int = columns
		set(value) {
			if (field != value) {
				field = value
				invalidateLayout()
			}
		}

	var rows: Int = rows
		set(value) {
			if (field != value) {
				field = value
				invalidateLayout()
			}
		}

	var gravity: RenderGravity = RenderGravity.CENTER_CENTER
		set(value) {
			if (field != value) {
				field = value
				invalidateLayout()
			}
		}

	var layout: Layout = Layout.TOP_LEFT
		set(value) {
			if (field != value) {
				field = value
				invalidateLayout()
			}
		}

	enum class Layout {
		TOP_LEFT {
			override fun get(
				list: List<EditablePanel<*>>,
				row: Int,
				column: Int,
				rows: Int,
				columns: Int
			): EditablePanel<*>? {
				return list.getOrNull(column + row * columns)
			}
		},

		TOP_RIGHT {
			override fun get(
				list: List<EditablePanel<*>>,
				row: Int,
				column: Int,
				rows: Int,
				columns: Int
			): EditablePanel<*>? {
				return list.getOrNull((columns - column - 1) + row * columns)
			}
		},

		BOTTOM_LEFT {
			override fun get(
				list: List<EditablePanel<*>>,
				row: Int,
				column: Int,
				rows: Int,
				columns: Int
			): EditablePanel<*>? {
				return list.getOrNull(column + (rows - row - 1) * columns)
			}
		},

		BOTTOM_RIGHT {
			override fun get(
				list: List<EditablePanel<*>>,
				row: Int,
				column: Int,
				rows: Int,
				columns: Int
			): EditablePanel<*>? {
				return list.getOrNull((columns - column - 1) + (rows - row - 1) * columns)
			}
		};

		abstract fun get(list: List<EditablePanel<*>>, row: Int, column: Int, rows: Int, columns: Int): EditablePanel<*>?
	}

	override fun sizeToContents() {
		if (visibleChildren.isEmpty()) {
			// nothing to size against
			return
		}

		visibleChildren.forEach { it.sizeToContents() }

		var width = 0f
		var height = 0f

		val children = visibleChildren.iterator().filter { it.dock == Dock.NONE }

		for (row in 0 until rows) {
			var maxHeight = 0f
			var rowWidth = 0f

			for (column in 0 until columns) {
				val nextChild = children.maybe() ?: break
				rowWidth += nextChild.width + nextChild.dockMargin.horizontal
				maxHeight = max(maxHeight, nextChild.height + nextChild.dockMargin.vertical)
			}

			height += maxHeight
			width = max(width, rowWidth)
		}

		this.width = width
		this.height = height
	}

	override fun performLayout() {
		super.performLayout()
		val children = visibleChildren.stream().filter { it.dock == Dock.NONE }.toList()

		var totalWidth = 0f
		var totalHeight = 0f

		for (row in 0 until rows) {
			var maxHeight = 0f
			var width = 0f

			for (column in 0 until columns) {
				val child = layout.get(children, row, column, rows, columns) ?: continue
				width += child.dockedWidth
				maxHeight = maxHeight.coerceAtLeast(child.dockedHeight)
			}

			totalWidth = totalWidth.coerceAtLeast(width)
			totalHeight += maxHeight
		}

		val alignX = gravity.repositionX(width, totalWidth)
		val alignY = gravity.repositionY(height, totalHeight)

		totalWidth = 0f
		totalHeight = 0f

		for (row in 0 until rows) {
			var maxHeight = 0f
			var width = 0f

			for (column in 0 until columns) {
				val child = layout.get(children, row, column, rows, columns) ?: continue
				child.x = alignX + width
				child.y = alignY + totalHeight
				width += child.dockedWidth
				maxHeight = maxHeight.coerceAtLeast(child.dockedHeight)
			}

			totalWidth = totalWidth.coerceAtLeast(width)
			totalHeight += maxHeight
		}
	}

	companion object {
		fun <S : Screen> slots(screen: S, parent: EditablePanel<*>?, columns: Int, rows: Int): GridPanel<S> {
			return GridPanel(screen, parent, width = columns * AbstractSlotPanel.SIZE, height = rows * AbstractSlotPanel.SIZE, rows = rows, columns = columns)
		}
	}
}
